from lxml import etree
from zeep import Plugin


class RemoveNamespacePlugin(Plugin):
    """Remove a namespace declaration and corresponding prefixes before sending off the XML to NAV.

    Useful when NAV complains about `The Element <ns1:XXX> is unexpected` (possibly fixable by using https://support.microsoft.com/en-hk/help/2509042/-the-element-is-unexpected-error-message-when-you-run-certain-xml-port)

    Args:
        namespace (str):
            The namespace to remove.
            E.g: 'urn:microsoft-dynamics-nav/xmlports/x50001'
    """

    def __init__(self, namespace):
        self.namespace = namespace

    def egress(self, envelope, http_headers, operation, binding_options):
        ns = u'{%s}' % self.namespace
        nsl = len(ns)

        for elem in envelope.getiterator():
            if elem.tag.startswith(ns):
                elem.tag = elem.tag[nsl:]

        etree.cleanup_namespaces(envelope)

        return envelope, http_headers



"""
Simple wrapper that adds some extra encoding capabilities needed for
this project.
"""
import collections
import datetime
import decimal
from json import JSONDecodeError  # noqa
import json as json_impl


class JsonExtendedEncoder(json_impl.JSONEncoder):
    """
    Needed for the json module to understand what to do with the types we
    use in this project.
    """
    def default(self, obj):
        if isinstance(obj, decimal.Decimal):
            return float(obj)
        elif isinstance(
            obj,
            (
                datetime.time,
                datetime.date,
                datetime.datetime,
            ),
        ):
            return str(obj)

        # Handle case where zeep returns the undocumented _raw_elements key
        # which is of type `collections.deque`. As the name suggestions this
        # object contains raw elements, which json will be unable to process,
        # therefore we return None.
        elif isinstance(obj, collections.deque):
            return str([x for x in obj])
        return super().default(obj)


def dump(*args, **kw):
    if 'cls' not in kw:
        kw['cls'] = JsonExtendedEncoder
    return json_impl.dump(*args, **kw)


def dumps(*args, **kw):
    if 'cls' not in kw:
        kw['cls'] = JsonExtendedEncoder
    return json_impl.dumps(*args, **kw)


load = json_impl.load
loads = json_impl.loads



import zeep.helpers

from . import constants

UNSET = object()


def convert_string_filter_values(filters):
    """Convert string type filter values to their XML equivalent
    Mostly useful for parsing CLI input data.
    """
    return {
        field: constants.STRING_VALUE_TO_PYTHON_TYPE_MAPPING.get(
            criteria,
            criteria
        )
        for field, criteria in filters.items()
    }


def to_builtins(data, default=UNSET, target_cls=dict):
    """
    Turn zeep XML object into python built-in data structures

    Args:
        default (Any):
            A default to return when serialized object is None.
        target_cls (Type[dict]):
            Which type of dictionary type will be used for objects.
            As this project's minimum Python version officially supported is
            3.6 we can rely on the native sorted order of the standard `dict`
            class as a default.
    """
    d = zeep.helpers.serialize_object(data)
    if d is None and default is not UNSET:
        return default
    return zeep.helpers.serialize_object(d, target_cls=target_cls)



from zeep.xsd.elements.element import NotSet

DEFAULT_WSDL_CACHE_EXPIRATION = 3600

CODEUNIT = 'Codeunit'
PAGE = 'Page'

ReadMultiple = 'ReadMultiple'
CreateMultiple = 'CreateMultiple'

STRING_VALUE_TO_PYTHON_TYPE_MAPPING = {
    'NotSet': NotSet,
    'TRUE': True,
    'FALSE': False,
}
